@@ -6,13 +6,13 @@ import multiprocessing
 import queue
 import subprocess
 from typing import Tuple, Dict, Callable, List, Union
+import pprint
 
 import eventlet  # type: ignore
 import flask_socketio  # type: ignore
 
 import pytest  # type: ignore
 from _pytest import reports  # type: ignore
-from py._path import local  # type: ignore
 
 from pytest_web_ui import result_tree
 
@@ -71,7 +71,7 @@ class PyTestRunner:
         result_node.report = report
 
     def _send_update(self, updated_node: result_tree.Node):
-        if not updated_node.nodeid:
+        if updated_node.nodeid == self.result_tree.nodeid:
             parents_slice = self._branch_schema.dump(updated_node)
         else:
             parents_slice, parent_node = result_tree.serialize_parents_slice(
@@ -93,9 +93,7 @@ class PyTestRunner:
         self._socketio.emit("update", parents_slice)
 
 
-def _run_test(
-    nodeid: str, queue: multiprocessing.Queue, root_dir: local.LocalPath, test_dir: str
-):
+def _run_test(nodeid: str, queue: multiprocessing.Queue, root_dir: str, test_dir: str):
     full_path = _get_full_path(nodeid, root_dir, test_dir)
     LOGGER.debug("full_path: %s", full_path)
     plugin = TestRunPlugin(queue=queue)
@@ -103,11 +101,10 @@ def _run_test(
     queue.put(_DONE)
 
 
-def _get_full_path(nodeid: str, root_dir: local.LocalPath, test_dir: str) -> str:
-    LOGGER.critical("typeof root_dir: %s", type(root_dir))
+def _get_full_path(nodeid: str, root_dir: str, test_dir: str) -> str:
     if not nodeid:
         return test_dir
-    return str(root_dir / nodeid.replace("/", os.sep))
+    return nodeid.replace("/", os.sep)
 
 
 class EnvironmentManager:
@@ -115,18 +112,20 @@ class EnvironmentManager:
     COMPOSE_FILENAME = "docker_compose.yml"
 
     def __init__(self, directory: str, enable: bool):
-        self._directory = directory
+        self._compose_path = os.path.join(directory, self.COMPOSE_FILENAME)
         self._proc = None
         self._enable = enable
 
     def __enter__(self):
-        compose_path = os.path.join(self._directory, self.COMPOSE_FILENAME)
-        if self._enable and os.path.exists(compose_path):
-            self._proc = subprocess.Popen(["docker-compose", "-f", compose_path, "up"])
+        if self._enable and os.path.exists(self._compose_path):
+            self._proc = subprocess.Popen(
+                ["docker-compose", "-f", self._compose_path, "up"]
+            )
         return self
 
     def __exit__(self, exc_type, exc_value, exc_tb):
         if self._proc:
+            subprocess.check_call(["docker-compose", "-f", self._compose_path, "down"])
             self._proc.wait()
 
 
@@ -134,13 +133,42 @@ def _init_result_tree(
     directory: str,
 ) -> Tuple[result_tree.BranchNode, Dict[str, result_tree.Node]]:
     """Collect the tests and initialise the result tree skeleton."""
-    plugin = CollectPlugin()
-    ret = pytest.main(["--collect-only", directory], plugins=[plugin])
-    if ret != 0:
-        raise RuntimeError(f"Failed to collect tests from {directory}")
+    node, index = _init_result_tree_recur(directory)
+    for child_branch in node.child_branches.values():
+        result_tree.set_parent_ids(child_branch)
+    return node, index
 
-    session = plugin.session
-    return result_tree.build_from_session(session, directory)
+
+def _init_result_tree_recur(
+    directory: str,
+) -> Tuple[result_tree.BranchNode, Dict[str, result_tree.Node]]:
+    root_node = result_tree.BranchNode(
+        nodeid=directory.replace(os.sep, "/"),
+        fspath=directory,
+        short_id=os.path.basename(directory),
+    )
+    nodes_index: Dict[str, result_tree.Node] = {root_node.nodeid: root_node}
+
+    with os.scandir(directory) as it:
+        for entry in it:
+            if entry.is_file() and entry.name.endswith(".py"):
+                plugin = CollectPlugin()
+                ret = pytest.main(["--collect-only", entry.path], plugins=[plugin])
+                if ret != 0:
+                    raise RuntimeError(f"Failed to collect tests from {entry.path}")
+
+                session = plugin.session
+                node, index = result_tree.build_from_session(session, entry.path)
+            elif entry.is_dir():
+                node, index = _init_result_tree_recur(entry.path)
+            else:
+                continue
+
+            if list(node.iter_children()):
+                root_node.child_branches[node.short_id] = node
+                nodes_index.update(index)
+
+    return root_node, nodes_index
 
 
 class CollectPlugin:
